/**
 * Copyright (C) 2005 Gregoire Lejeune <gregoire.lejeune@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include "xslt.h"

/**
 * external function support, patch from :
 *
 * Brendan Taylor 
 * whateley@gmail.com
 *
 */

/* this is needed to free the xmlHashTablePtr */
void deallocateCallback( void *payload, xmlChar *name ATTRIBUTE_UNUSED ) {
  TMethodCall *method = (TMethodCall *)payload;

  free(method);
}

/* 
 * number -> float
 * boolean -> boolean
 * string -> string
 * nodeset -> an array of strings
 */
VALUE xpathObj2value(xmlXPathObjectPtr obj, xmlDocPtr doc)
{
  VALUE ret;

  if (obj == NULL) {
    return Qnil;
  }

  switch (obj->type) {
    case XPATH_XSLT_TREE: { 
      /* XXX i don't know if this works because i don't know what this type is */
      if ((obj->nodesetval == NULL) || (obj->nodesetval->nodeNr == 0) ||
                      (obj->nodesetval->nodeTab == NULL)) {
        ret = rb_ary_new();
      } else {
        xmlBufferPtr buff = xmlBufferCreate();
        xmlNodePtr node;

        node = obj->nodesetval->nodeTab[0]->children;
        ret = rb_ary_new();
        while (node != NULL) {
          xmlNodeDump(buff,doc,node,0,0);
          rb_ary_push(ret, rb_str_new2((char*) buff->content));
          node = node->next;
        }
      }
      return (ret);
    }
    case XPATH_NODESET:
      ret = rb_ary_new();
      
      if ((obj->nodesetval != NULL) && (obj->nodesetval->nodeNr != 0)) {
        xmlBufferPtr buff = xmlBufferCreate();
        xmlNodePtr node;
        int i;

        /* XXX these should be a different type on a node-by-node basis */
        for (i = 0; i < obj->nodesetval->nodeNr; i++) {
          node = obj->nodesetval->nodeTab[i];
          xmlNodeDump(buff, doc, node, 0, 0);
          rb_ary_push(ret, rb_str_new2(buff->content)); 
        }
      }
      break;
    case XPATH_BOOLEAN:
      ret = obj->boolval ? Qtrue : Qfalse;
      break;
    case XPATH_NUMBER:
      ret = rb_float_new(obj->floatval);
      break;
    case XPATH_STRING:
      ret = rb_str_new2((char *) obj->stringval);
      break;
    case XPATH_POINT:
    case XPATH_RANGE:
    case XPATH_LOCATIONSET:
    default:
      printf("Unable to convert XPath object type %d\n", obj->type);
      ret = Qnil;
  }
  xmlXPathFreeObject(obj);
  return ret;
}

/*
 * boolean -> boolean
 * number -> number
 * string -> parsed node or string
 * array -> array of parsed nodes
 */
xmlXPathObjectPtr value2xpathObj (VALUE val) {
  xmlXPathObjectPtr ret = NULL;
  
  switch (TYPE(val)) {
    case T_TRUE:
    case T_FALSE:
      ret = xmlXPathNewBoolean(RTEST(val));
      break;
    case T_FIXNUM:
    case T_FLOAT:
      ret = xmlXPathNewFloat(NUM2DBL(val));
      break;
    case T_STRING:
    {
      /* attempt to parse it as a document. 
       * if it's illformed, just use it as a string */
      /* XXX there must be a better way, 
       * this prints an annoying message on failure */
      xmlDocPtr doc = xmlParseDoc((xmlChar *) STR2CSTR(val));
  
      if( doc != NULL ) {
        ret = xmlXPathNewNodeSet((xmlNode *)doc->children);
      } else {
        ret = xmlXPathWrapString(STR2CSTR(val));
      }
      break;
    }
    case T_ARRAY:
    {
      xmlDocPtr doc; 
      xmlNodeSetPtr set;
      int i;

      set = xmlXPathNodeSetCreate(NULL);

      for(i = RARRAY(val)->len; i > 0; i--) {
        doc = xmlParseDoc((xmlChar *) STR2CSTR(rb_ary_shift(val)));

      /* XXX this is bad. it just drops things that aren't well formed */
        if( doc != NULL ) {
          xmlXPathNodeSetAdd(set, (xmlNode *)doc->children);
        }
      }

      ret = xmlXPathWrapNodeSet(set);
      break;
      /* should probably also have a case for Qnil */
    }
    default:
      printf("value2xpathObj: can't convert an object of class %s to XPath\n", getRubyObjectName(val));
      return NULL;
  }

  return ret;
}

void xmlXPathFuncCallback( xmlXPathParserContextPtr ctxt, int nargs) {
  VALUE result, arguments[nargs];
  const xmlChar *name, *namespace;
  xmlXPathObjectPtr obj;
  int i;

  TMethodCall *mycall = NULL;

  if (ctxt == NULL || ctxt->context == NULL)
    return;

  name = ctxt->context->function;
  namespace = ctxt->context->functionURI;

  mycall = xmlHashLookup2(pExtFunctionsHashTable, name, namespace);

  if (mycall == NULL) {
    printf("xmlXPathFuncCallback: internal error %s not found !\n", name);
    return;
  }

  /* arguments come from the context in reverse order */
  for (i = nargs - 1; i >= 0; i--) {
    obj = valuePop(ctxt);
    arguments[i] = xpathObj2value(obj, ctxt->context->doc);
  }

  result = rb_funcall2( mycall->object, mycall->method, nargs, arguments);

  obj = value2xpathObj(result);
  valuePush(ctxt, obj);
}

/**
 * ----------------------------------------------------------------------------
 */

/**
 * external function support, path from :
 *
 * Brendan Taylor 
 * whateley@gmail.com
 *
 */
VALUE ruby_xslt_ext_function( VALUE self, VALUE name, VALUE namespace, VALUE object, VALUE method ) {
  RbTxslt *pRbTxslt;
  TMethodCall *mycall = (TMethodCall *)malloc(sizeof(TMethodCall));

  Data_Get_Struct( self, RbTxslt, pRbTxslt );

  if( pExtFunctionsHashTable == NULL )
    pExtFunctionsHashTable = xmlHashCreate( 10 );
  
  if( pExtFunctionsHashTable == NULL )
	rb_raise( rb_eNoMemError, "Could not allocate hash for external functions" );
  
  mycall->object = object;
  mycall->method = rb_intern(STR2CSTR(method));

  xmlHashAddEntry2( pExtFunctionsHashTable, STR2CSTR(name), STR2CSTR(namespace), mycall );

  xsltRegisterExtModuleFunction( STR2CSTR(name), STR2CSTR(namespace), xmlXPathFuncCallback );

  pRbTxslt->iXmlResultType = RUBY_XSLT_XMLSRC_TYPE_NULL;
  
  /* this needs real return values probably */
  return Qnil;
}

